InterpreterPlugin subclass: #AioPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcess'!

!AioPlugin commentStamp: '<historical>' prior: 0!
AioPlugin provides a primitive interface to asynchronous input/output events. Different
classes are required for various operating system platforms. All implementations are expected
to use the module name "AioPlugin" regardless of their class name. This provides a uniform
primitive interface for asynchronous events on all supported platforms.

Aio events are externally occuring events such as "data is now available on an IO channel"
or "an IO channel is now available for writing". These events occur asynchronously with
respect to the Squeak VM. AioPlugin arranges for event forwarding such that each event
causes a corresponding Smalltalk semaphore to be signalled.

To create an event handler in Squeak, register an external semaphore, establish event
forwarding to the semaphore, and start a Smalltalk process to wait on the semaphore. The
process should loop on the semaphore wait, taking appropriate action when the event is
signalled, and terminating when the event forwarding is disabled.

The supported primitives are:
  primitiveAioEnable - Enable asynchronous notification for a descriptor.
  primitiveAioHandle - Handle asynchronous event notification for a descriptor.
  primitiveAioSuspend - Temporarily suspend asynchronous event notification for a descriptor.
  primitiveAioDisable - Definitively disable asynchronous event notification for a descriptor.
  primitiveOSFileHandle - Answer the handle (Unix file number) for a FileStream.
  primitive OSSocketHandle - Answer the handle (Unix file number) for a Socket.

Currently the Unix VM supports aio event notification, and class UnixAioPlugin provides
the required primitive support.
!

AioPlugin subclass: #UnixAioPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcess'!

!UnixAioPlugin commentStamp: '<historical>' prior: 0!
UnixAioPlugin implements the Unix platform methods for AioPlugin.!


!AioPlugin methodsFor: 'initialize-release' stamp: 'dtl 9/1/2003 19:55'!
initialiseModule

	self export: true
! !

!AioPlugin methodsFor: 'initialize-release' stamp: 'dtl 9/1/2003 19:55'!
shutdownModule

	self export: true
! !

!AioPlugin methodsFor: 'primitives' stamp: 'dtl 3/12/2005 22:50'!
primitiveModuleName
	"Answer a string containing the module name string for this plugin."

	| s |
	self export: true.
	self var: 's' declareC: 'char *s'.
	self cCode: 's= (char *)moduleName' inSmalltalk: [s _ moduleName]. "Prevent compiler warning"
	interpreterProxy pop: 1; push: (self stringFromCString: s)
! !

!AioPlugin methodsFor: 'primitives' stamp: 'dtl 9/1/2003 20:11'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	self export: true.
	interpreterProxy pop: 1 thenPush: (self stringFromCString: self versionString)
! !

!AioPlugin methodsFor: 'version identification' stamp: 'dtl 9/7/2005 06:12'!
versionString
	"Answer a string containing the version string for this plugin. Handle MNU
	errors, which can occur if class InterpreterPlugin has been removed from
	the system.

	Important: When this method is changed, the class side method must also be
	changed to match."

	"2.0 supports 64bit code base"

	| version |
	self returnTypeC: 'char *'.
	self var: 'version' declareC: 'static char version[]= "2.0"'.
	^ self cCode: 'version' inSmalltalk: [self class versionString]
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 7/4/2003 19:23'!
fileDescriptorFrom: aFileHandle
	"Answer the OS file descriptor, an integer value, from a SQFile data structure,
	or answer -1 if unable to obtain the file descriptor (probably due to receiving
	an incorrect type of object as aFileHandle). This method may be called from a
	primitive, and is not intended to be called from Smalltalk."

	| sqFile osFileStream |
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'osFileStream' declareC: 'FILE *osFileStream'.
	sqFile _ self fileValueOf: aFileHandle.
	sqFile = 0
		ifTrue:
			[^ -1]	"Something is wrong, bail out before we core dump."
		ifFalse:
			[osFileStream _ self cCode: 'sqFile->file'.
			osFileStream = 0
				ifTrue:
					[^ -1].	"Something is wrong, bail out before we core dump."
			^ self fileno: osFileStream]
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 7/4/2003 19:21'!
fileRecordSize
	"Answer the size of a SQFile data structure in bytes."

	^ self cCode: 'sizeof(SQFile)' inSmalltalk: [20]
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 7/4/2003 19:22'!
fileValueOf: anSQFileRecord
	"Return a pointer to the first byte of of the SQFile data structure file record within
	anSQFileRecord, which is expected to be a ByteArray of size self>>fileRecordSize."

	self returnTypeC: 'SQFile *'.
	((interpreterProxy isBytes: anSQFileRecord) and: [
		(interpreterProxy byteSizeOf: anSQFileRecord) == self fileRecordSize])
			ifTrue:
				[^ interpreterProxy arrayValueOf: anSQFileRecord]
			ifFalse:
				[interpreterProxy primitiveFail.
				^ nil]

! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 7/4/2003 19:22'!
isSQFileObject: objectPointer
	"Answer true if objectPointer appears to be a valid SQFile ByteArray. This check
	is appropriate if objectPointer has been passed as a parameter to a primitive, and
	is expected to represent a valid file reference."

	| idx sqFileBytes |
	self var: 'sqFileBytes' declareC: 'unsigned char *sqFileBytes'.
	(((interpreterProxy isBytes: objectPointer) and:
		 [(interpreterProxy byteSizeOf: objectPointer) = self fileRecordSize]))
			ifFalse: [^ false].
	"Check for the common failure mode of a SQFile record with all zeros."
	sqFileBytes _ interpreterProxy arrayValueOf: objectPointer.
	idx _ 0.
	[idx < self fileRecordSize]
		whileTrue:
			[((sqFileBytes at: idx) ~= 0) ifTrue: [^ true].
			idx _ idx + 1].
	^ false
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 7/4/2003 19:23'!
isSQSocketObject: objectPointer
	"Answer true if objectPointer appears to be a valid SQSocket ByteArray. This check
	is appropriate if objectPointer has been passed as a parameter to a primitive, and
	is expected to represent a valid socket reference."

	| idx sqSocketBytes |
	self var: 'sqSocketBytes' declareC: 'unsigned char *sqSocketBytes'.
	(((interpreterProxy isBytes: objectPointer) and:
		 [(interpreterProxy byteSizeOf: objectPointer) = self socketRecordSize]))
			ifFalse: [^ false].
	"Check for the common failure mode of a SQSocket record with all zeros."
	sqSocketBytes _ interpreterProxy arrayValueOf: objectPointer.
	idx _ 0.
	[idx < self socketRecordSize]
		whileTrue:
			[((sqSocketBytes at: idx) ~= 0) ifTrue: [^ true].
			idx _ idx + 1].
	^ false
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 7/4/2003 19:24'!
socketDescriptorFrom: sqSocketOop
	"Answer the OS file descriptor, an integer value, from a SQSocket data structure,
	or answer -1 if unable to obtain the file descriptor (probably due to receiving
	an incorrect type of object as aFileHandle)."

	"Gross hack warning: The first element of privateSocketStruct happens to be
	the Unix file number of the socket. See sqUnixSocket.c for the definition. This
	method is coded to take advantage of this factoid, and will break if anyone
	ever redefines the data structure."

	| sqSocket privateSocketStruct descriptor |
	self var: 'sqSocket' declareC: 'SQSocket *sqSocket'.
	self var: 'privateSocketStruct' declareC: 'void *privateSocketStruct'.
	sqSocket _ self socketValueOf: sqSocketOop.
	sqSocket = 0
		ifTrue:
			[^ -1]	"Something is wrong, bail out before we core dump."
		ifFalse:
			[privateSocketStruct _ self cCode: 'sqSocket->privateSocketPtr'.
			privateSocketStruct = 0
				ifTrue:
					[^ -1].	"Something is wrong, bail out before we core dump."
			"Gross hack occurs here. FIXME"
			descriptor _ self cCode: '* (int *) privateSocketStruct' inSmalltalk: [-1].
			^ descriptor]
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 7/4/2003 19:24'!
socketRecordSize
	"Answer the size of a SQSocket data structure in bytes."

	^ self cCode: 'sizeof(SQSocket)' inSmalltalk: [12]
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 7/4/2003 19:24'!
socketValueOf: anSQSocketRecord
	"Return a pointer to the first byte of of the SQsocket data structure socket record within
	anSQSocketRecord, which is expected to be a ByteArray of size self>>socketRecordSize."

	self returnTypeC: 'SQSocket *'.
	((interpreterProxy isBytes: anSQSocketRecord) and: [
		(interpreterProxy byteSizeOf: anSQSocketRecord) == self socketRecordSize])
			ifTrue: [^ interpreterProxy arrayValueOf: anSQSocketRecord]
			ifFalse: [^ nil]

! !

!AioPlugin methodsFor: 'private' stamp: 'dtl 9/1/2003 17:12'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString stringPtr |
	self var: 'aCString' declareC: 'char *aCString'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	len _ self cCode: 'strlen(aCString)' inSmalltalk: [0].
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	stringPtr _ interpreterProxy arrayValueOf: newString.
	self cCode: '(char *)strncpy(stringPtr, aCString, len)'.
	^ newString

! !


!AioPlugin class methodsFor: 'translation' stamp: 'dtl 9/17/2005 10:36'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '"FilePlugin.h"'.
	cg addHeaderFile: '"SocketPlugin.h"'.
	cg addHeaderFile: '"config.h"
#ifndef SQAIO_H
# define SQAIO_H "aio.h"  /* aio.h has been renamed to sqaio.h */
#endif'.
	cg addHeaderFile: 'SQAIO_H'
! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:17'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ false! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:17'!
moduleName

	^ 'AioPlugin'! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:17'!
shouldBeTranslated

	^ self isResponsibleForThisPlatform
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 9/1/2003 17:23'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"self installedModuleName"

	<primitive: 'primitiveModuleName' module: 'AioPlugin'>
	^ nil
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 9/1/2003 17:22'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"self installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'AioPlugin'>
	^ nil
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 9/7/2005 05:44'!
versionString
	"Important: When this method is changed, the instance side method must also be
	changed to match."

	"self versionString"

	"2.0 supports 64bit code base"

	^ '2.0'! !


!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveAioDisable
	"Definitively disable asynchronous event notification for a descriptor. The
	parameter is an OS level integer file descriptor."

	| fd |
	self export: true.
	fd _ interpreterProxy stackIntegerValue: 0.
	(fd == interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	self aioDisable: fd.
	interpreterProxy pop: 2; pushInteger: fd
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveAioEnable
	"Enable asynchronous notification for a descriptor. The first parameter is an OS
	level integer file descriptor. The second parameter is the index of a Semaphore to
	be notified, and the third parameter is a flag indicating that descriptor represents
	an external object and should not be closed on termination of aio handling. Answer
	the semaphore index."

	| fd semaIndex flags externalObject eventSemaphoreIndices |
	self export: true.
	self var: 'eventSemaphoreIndices' declareC: 'static int eventSemaphoreIndices[FD_SETSIZE]'.
	fd _ interpreterProxy stackIntegerValue: 2.
	(fd == interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	semaIndex _ interpreterProxy stackIntegerValue: 1.
	"The following is to prevent the need to coerce integers into pointers in clientData".
	eventSemaphoreIndices at: semaIndex put: semaIndex.
	externalObject _ interpreterProxy stackObjectValue: 0.
	(externalObject = interpreterProxy trueObject)
		ifTrue: [flags _ self cCode: 'AIO_EXT' inSmalltalk: [16]]
		ifFalse: [flags _ 0].
	self cCode: 'aioEnable(fd, &(eventSemaphoreIndices[semaIndex]), flags)'.
	interpreterProxy pop: 4; pushInteger: semaIndex.! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveAioHandle
	"Handle asynchronous event notification for a descriptor. The first parameter is
	an OS level integer file descriptor. The remaining three parameters are Boolean
	flags representing the types of events for which notification is being requested:
	handle exceptions, handle for read, and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| fd exceptionWatch readWatch writeWatch flags |
	self export: true.
	fd _ interpreterProxy stackIntegerValue: 3.
	(fd == interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch _ interpreterProxy stackObjectValue: 2.
	readWatch _ interpreterProxy stackObjectValue: 1.
	writeWatch _ interpreterProxy stackObjectValue: 0.
	flags _ 0.
	(exceptionWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self cCode: 'aioHandle(fd, aioForwardwithDataandFlags, flags)'.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveAioSuspend
	"Temporarily suspend asynchronous event notification for a descriptor. The first
	parameter is an OS level integer file descriptor. The remaining three parameters
	are Boolean flags representing the types of events for which notification is being
	requested: handle exceptions, handle for read, and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| fd exceptionWatch readWatch writeWatch flags |
	self export: true.
	fd _ interpreterProxy stackIntegerValue: 3.
	(fd == interpreterProxy nilObject) ifTrue: [^ interpreterProxy primitiveFail].
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch _ interpreterProxy stackObjectValue: 2.
	readWatch _ interpreterProxy stackObjectValue: 1.
	writeWatch _ interpreterProxy stackObjectValue: 0.
	flags _ 0.
	(exceptionWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self cCode: 'aioSuspend(fd, flags)'.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveOSFileHandle
	"Take a struct SQFile from the stack, and answer the value of its Unix file number."

	| sqFileOop fileNo |
	self export: true.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo _ self fileDescriptorFrom: sqFileOop.
	interpreterProxy pop: 2; pushInteger: fileNo
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl 7/4/2003 19:18'!
primitiveOSSocketHandle
	"Take a struct SQSocket from the stack, and answer the value of its Unix file number."

	| sqSocketOop fileNo |
	self export: true.
	sqSocketOop _ interpreterProxy stackValue: 0.
	(self isSQSocketObject: sqSocketOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo _ self socketDescriptorFrom: sqSocketOop.
	(fileNo < 0) ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 2; pushInteger: fileNo
! !

!UnixAioPlugin methodsFor: 'private - event forwarding' stamp: 'dtl 9/13/2005 22:03'!
aioForward: fd withData: data andFlags: flags
	"This function is called to signal a Smalltalk Semaphore when an asynchronous event is
	detected. When translated to C, the name of this method is aioForwardwithDataandFlags.
	The event handler is set up by #primitiveAioHandle."

	| semaIndex pfd |
	self returnTypeC: 'void'.
	self var: 'fd' type: 'int'.
	self var: 'data' type: 'void *'.
	self var: 'flags' type: 'int'.
	self var: 'pfd' declareC: 'int *pfd'.
	pfd _ data.
	semaIndex _ self cCode: '*pfd' inSmalltalk: [0].
	interpreterProxy signalSemaphoreWithIndex: semaIndex
! !


!UnixAioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:35'!
declareCVarsIn: cg
! !

!UnixAioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:18'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ Smalltalk platformName = 'unix'! !

